{% load pages_tags mezzanine_tags i18n future %}
{% nevercache %}
{% for message in messages %}
<div class="alert alert-{{ message.tags }}" data-alert="alert">
<a class="close" href="#" onclick="$(this).parent().fadeOut('fast'); return false;">×</a>{{ message }}
</div>
{% endfor %}
{% endnevercache %}
