{% load settings_tags %}
<div class="container-fluid" style="background:ffd;">
  {% include "SideBar.html" %}
   <div class="content">
     {% include "HeroUnit.html" %}
     {% include "ThreeColumns.html" %}
     {% include "FooterSimple.html" %}
   </div>
</div>
