 {% include "MezzanineNavBar.html" %}
{# {% include "Bootstrap3NavBar.html" %} #}

{% comment %}
<div style="text-align:center;color:#fff;height:25px;width:100%;background:#88f;">
This is the top bar, it is named "TopBar.html" If you place it outside the 'container' it will fill the whole width.
</div>
{% endcomment %}
