{% comment %}


<!--<link rel="stylesheet" href="{{ STATIC_URL }}styles/bootstrap24.min.css">-->
<link rel="stylesheet" href="{{ STATIC_URL }}css/bootstrap.css">
<!--<link rel="stylesheet" href="http://66.113.100.140:81/apps/bootstrap3/css/bootstrap.css">-->
{% if csettings.Mezzanine == True %}
  {% include "MezzanineCSS.html" %}
{% endif %}


{% endcomment %}