INSTALLED_APPS = [
   'django.contrib.auth',
   'django.contrib.contenttypes',
   'grappelli_safe', #must precede admin
   'django.contrib.redirects',
   'django.contrib.sessions',
   'django.contrib.sites',
   'django.contrib.sitemaps',
   'django.contrib.humanize',
   'django.contrib.comments',
   'django.contrib.admin',
   'django.contrib.staticfiles',
   'filebrowser_safe',
   'south',
   'debug_toolbar',
   'django_extensions',
   'compressor',
   'sekizai',
   'selectable',
   'backstage',
]


#   'django.contrib.markup',

#   'django.contrib.flatpages',
#   'django.contrib.staticfiles',
